/* PATTERN.C - bydate pattern manipulation functions */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <environ.h>
#include "pattern.h"

/*
   These defines list and name the valid pattern element ids
*/
#define CENTURY      'c'
#define YEAR         'y'
#define MONTH        'm'
#define DAY          'd'
#define HOUR         'H'
#define MINUTE       'M'
#define SECOND       'S'
#define PREFIX       'p'
#define SUFFIX       's'
#define FILENAME     'f'
#define TEXT         't'
#define EXTENSION    'e'

int         dbg = 0;

/*
   Private function prototypes
*/
void clear_error(PCB *p);
void dump_pattern(PCB *p);
void set_error_code(PCB *p, char *fmt, char c);
void set_error_num(PCB *p, char e, char *s, int n);
void set_error_text(PCB *p, char e, char *s, char c);
void set_sys_error(PCB *p, char *s, char *e);

/*
   parse_pattern() - parse a pattern, populating its decoded structure.
   ===============   Returns a pointer to the structure. The error,
                     if any, is stored in the PCB structure.
*/
PCB *parse_pattern(char *pat, int test_mode, int debug)
{
   PCB         *p = NULL;
   ELEMENT     *e = NULL;
   int         i;
   int         j;
   int         l;
   int         n = strlen(pat);
   int         o;
   char        c;
   int         stop;
   time_t      now;
   struct tm   *test_time;
   
   dbg = debug;

   /*
      Initialise the pattern control block
      and decode the pattern to populate it
   */
   p = (PCB*)get_space(sizeof(PCB));
   p->pattern = (char*)get_space(n + 1);
   strcpy(p->pattern, pat);
   p->error = NULL;
   p->date = NULL;
   p->elements = 0;
   p->e = (ELEMENT**)get_space(sizeof(ELEMENT*) * n);
   stop = FALSE;
   for (i = 0; i < n && !stop; i++)
   {
      c = p->pattern[i];
      e = (ELEMENT*)get_space(sizeof(ELEMENT));
      e->code = c;
      e->offset = 0;
      e->length = 0;
      e->value = 0;
      p->e[p->elements] = e;
      p->elements++;
      switch(c)
      {
         case CENTURY:
         case YEAR:
         case MONTH:
         case DAY:
         case HOUR:
         case MINUTE:
         case SECOND:
               /*
                  Only the element code is needed
               */
               break;
               
         case PREFIX:
         case SUFFIX:
               /*
                  The element is immediately followed by a length.
                  It may be absent or zero for PREFIX
               */
               l = 0;
               c = p->pattern[i + 1];
               while (isdigit(c))
               {
                  l = l * 10 + (c - '0');
                  c = p->pattern[++i + 1];
               }

               e->length = l;
               if (l > 255 || ( l < 1 && e->code == SUFFIX))
               {
                  set_error_num(p,
                                e->code,
                                "length is out of range:",
                                l);
                  stop = TRUE;
               }
               
               break;
               
         case FILENAME:
               /*
                  The element code is followed by (offset,length)
               */
               c = p->pattern[i + 1];
               o = 0;
               l = 0;
               if (c != '(')
               {
                  set_error_text(p,
                                 e->code,
                                 "not followed by",
                                 '(');
                  stop = TRUE;
               }
               else
               {
                  i += 2;
                  c = p->pattern[i];
                  while(c != ',' && i < n && c >= '0' && c <= '9')
                  {
                     o = o * 10 + (c - '0');
                     i++;
                     c = p->pattern[i];
                  }

                  o--;  /* correct offset (input starting from 1) */
                  if (i >= n || c != ',')
                  {
                     set_error_text(p,
                                    e->code,
                                    "position not terminated by",
                                    ',');
                     stop = TRUE;
                  }
                  else
                  {
                     i++;
                     c = p->pattern[i];
                     while(c != ')' && i < n && c >= '0' && c <= '9')
                     {
                        l = l * 10 + (c - '0');
                        i++;
                        c = p->pattern[i];
                     }

                     if (i >= n || c != ')')
                     {
                        set_error_text(p,
                                       e->code,
                                       "value not terminated by",
                                       ')');
                        stop = TRUE;
                     }
                  }
                  
                  e->offset = o;
                  e->length = l;

                  if (o + 1 < 1 || o + 1 > 255)
                  {
                     set_error_num(p,
                                   e->code,
                                   "offset is out of range:",
                                   o + 1);
                     stop = TRUE;
                  }

                  if (l < 1 || l > 255)
                  {
                     set_error_num(p,
                                   e->code,
                                   "length is out of range:",
                                   l);
                     stop = TRUE;
                  }
               }
               
               break;
               
         case TEXT:
         case EXTENSION:
               /*
                  The element code is followed by (string)
               */
               e->value = (char*)get_space(n+1);
               j = 0;
               c = p->pattern[i + 1];
               if (c != '(')
               {
                  set_error_text(p,
                                 e->code,
                                 "not followed by",
                                 '(');
                  stop = TRUE;
               }
               else
               {
                  i += 2;
                  c = p->pattern[i];
                  while(c != ')' && c != 0 && i < n)
                  {
                     e->value[j] = c;
                     i++;
                     j++;
                     c = p->pattern[i];
                  }

                  /*
                     Terminate the stored value and check that
                     a terminator was found.
                  */
                  e->value[j] = 0;
                  if (i >= n)
                  {
                     set_error_text(p,
                                    e->code,
                                    "value not terminated by",
                                    ')');
                     stop = TRUE;
                  }
               }

               /*
                  The extension element must be the last in the pattern.
                  Omit this test if an error has already been detected
                  so we don't overwrite it.
               */
               if (e->code == EXTENSION)
               {
                  if (i < n - 1 && !stop)
                  {
                     set_error_text(p,
                                    e->code,
                                    "is not the last element",
                                    ' ');
                     stop = TRUE;
                  }
               }
               
               break;
               
         default:
               set_error_code(p, "unknown element code: %c", c);
               stop = TRUE;
      }
   }

   /*
      Obtain the local time and date and store it in the control block
   */
   if (!stop)
   {
      if (!test_mode)
         now = time(0);
      else
      {
         test_time = get_space(sizeof(struct tm));

#ifdef MSDOS
         /*
            10:15:30 on 04 June 2009
         */
         test_time->tm_sec = 30;
         test_time->tm_min = 15;
         test_time->tm_hour = 10;

         test_time->tm_mday = 4;
         test_time->tm_mon = 5;
         test_time->tm_year = 109;

         test_time->tm_wday = 0; /* 4 */
         test_time->tm_yday = 0; /* 155 */
         test_time->tm_isdst = 0; /* 1 */
#else
         strptime("04/06/2009 09:15:30", "%d/%m/%C%y %H:%M:%S", test_time);
#endif
         now = mktime(test_time);
         if (debug)
            fprintf(stderr,
                    "Test mode using fixed date & time: %s\n",
                    asctime(test_time));

         free(test_time);
      }
      
      if (now == -1)
         set_sys_error(p, "Unable to get the date", strerror(errno));
      else
      {
         p->date = localtime(&now);
         if(!p->date)
            set_sys_error(p, "Invalid date found", strerror(errno));
      }
   }

   if (dbg)
      dump_pattern(p);
      
   return p;
}

/*
   free_pattern() - release the storage containing a pattern.
   ==============   Returns null;
*/
PCB *free_pattern(PCB *p)
{
   int      i;
   ELEMENT  *e;

   p->date = 0;
   for (i = 0; i < p->elements; i++)
   {
      e = p->e[i];
      if (e->value)
         release_space(e->value);
         
      release_space(p->e[i]);
   }

   release_space(p->e);
   clear_error(p);
   release_space(p->pattern);      
   release_space(p);
      
   if (dbg > 1)
      fprintf(stderr, "free_pattern()\n");
   
   return NULL;
}

/*
   pattern_ok() - test if the pattern contains an error.
   ============   Returns TRUE if the pattern is error free.
*/
int pattern_ok(PCB *p)
{
   int   r = (p->error == NULL);
   if (dbg > 1)
      fprintf(stderr, "%d = pattern_ok(%p)\n", r, p);
   return r;
}

/*
   pattern_error() - return the error associated with a pattern.
   ===============
*/
char *pattern_error(PCB *p)
{
   return p->error;
}

/*
   apply_pattern() - apply the pattern to the filename, creating
   ===============   the new file name, which is returned.
*/
char *apply_pattern(PCB *p, char *filename)
{
   char        *newname = NULL;
   int         i;
   int         j;
   int         k;
   int         n;
   int         w;
   int         lth;
   ELEMENT     *e;
   int         ext_element;
   char        *ext;
   char        buff[256];
   int         stop;

   if (dbg > 2)
      fprintf(stderr, "apply_pattern(%s,%s)\n", p->pattern, filename);
      
   /*
      Clear the error flag.
      Calculate the length of the new name.
   */
   clear_error(p);
   lth = 2;             /* Allow space for the '.' and terminator */
   ext_element = FALSE;
   stop = FALSE;
   for (i = 0; i < p->elements && !stop; i++)
   {
      e = p->e[i];
      switch(e->code)
      {
         case CENTURY:
         case YEAR:
         case MONTH:
         case DAY:
         case HOUR:
         case MINUTE:
         case SECOND:
                  lth += 2;                  /* all date and time elements */
                  break;                     /* are 2 chars                */
                        
         case PREFIX:
         case SUFFIX:
         case FILENAME:
                  lth += strlen(filename);   /* This element is never    */
                  break;                     /* longer than the filename */
               
         case TEXT:
                  lth += strlen(e->value);   /* Fixed length string */
                  break;

         case EXTENSION:
                  ext_element = TRUE;        /* Say whether element present */
                  lth += strlen(e->value);   /* Fixed length string */
                  break;

         default:
                  set_error_code(p,
                                 "is an unknown pattern CODE: 0x%02x",
                                 e->code);
                  stop = TRUE;
      }
   }

   if (!stop)
   {
      if (!ext_element)            /* if no override, the extension */
         lth += strlen(filename);  /* is not longer than filename   */
      
      newname = (char*)get_space(lth + 1);
      strcpy(newname, "");

      /*
         Process the pattern. Each time round the loop the next element
         in the pattern is processed, working left to right. Each element
         places its contribution to the file name in a string buffer which is
         appended to the new file name at the end of each iteration.
      */
      ext_element = FALSE;
      for (i = 0; i < p->elements; i++)
      {
         e = p->e[i];
         switch(e->code)
         {
            case CENTURY:  w = p->date->tm_year + 1900;
                           w = w / 100;
                           sprintf(buff, "%02d", w);
                           break;
                        
            case YEAR:     w = p->date->tm_year + 1900;
                           w = w % 100;
                           sprintf(buff, "%02d", w);
                           break;
                        
            case MONTH:    sprintf(buff, "%02d", p->date->tm_mon + 1);
                           break;
                        
            case DAY:      sprintf(buff, "%02d", p->date->tm_mday);
                           break;
                        
            case HOUR:     sprintf(buff, "%02d", p->date->tm_hour);
                           break;
                        
            case MINUTE:   sprintf(buff, "%02d", p->date->tm_min);
                           break;
                        
            case SECOND:   sprintf(buff, "%02d", p->date->tm_sec);
                           break;
                        
            case PREFIX:   strcpy(buff, filename);
                           ext = strchr(buff, '.');
                           if (ext)
                              *ext = 0;
                           
                           if (e->length >= 1)
                           {
                             if (e->length <= strlen(buff))
                                 buff[e->length] = 0;
                              else
                              {
                                 strcpy(buff, "");
                                 set_error_code(p,
                                                "%c is longer than filename",
                                                e->code);
                                 stop = TRUE;
                              }
                           }

                           break;
                           
            case SUFFIX:   strcpy(buff, filename);
                           ext = strchr(buff, '.');
                           if (ext)
                              *ext = 0;

                           n = strlen(buff);
                           if (e->length > n)
                           {
                              strcpy(buff, "");
                              set_error_code(p,
                                             "%c is longer than filename",
                                             e->code);
                              stop = TRUE;
                           }
                           else
                           {
                              for (j = 0, k = n - e->length; k < n; j++, k++)
                                 buff[j] = buff[k];

                              buff[j] = 0;
                           }

                           break;
               
            case FILENAME: strcpy(buff, filename);
                           ext = strchr(buff, '.');
                           if (ext)
                              *ext = 0;
                           
                           n = e->offset + e->length;
                       
                           if (n > strlen(buff))
                           {
                              strcpy(buff, "");
                              set_error_code(p,
                                             "%c won't fit in name",
                                             e->code);
                              stop = TRUE;
                           }
                           else
                           {
                              for (j = 0,k = e->offset; j < e->length; j++,k++)
                                 buff[j] = buff[k];

                              buff[j] = 0;
                           }
                        
                           break;
               
            case TEXT:     strcpy(buff, e->value);
                           break;
            
            case EXTENSION:
                           strcpy(buff, ".");
                           strcat(buff, e->value);
                           ext_element = TRUE;
                           break;
               
         }

         strcat(newname, buff);

         if (dbg > 2)
            fprintf(stderr, "%d: %c [%s]-> %s\n", i, e->code, buff, newname);
      }   

      /*
         Obtain and output the extension unless it was overridden
         by an EXTENSION element in the pattern. If there's no '.'
         in the filename than there's no extension to copy.
      */
      if (!ext_element)
      {
         ext = strchr(filename, '.');
         if (ext)
            strcat(newname, ext);
      }
   }
 
   /*
      All done: output diagnostics and return the result.
   */
   if (dbg > 1)
      fprintf(stderr,
              "%s = apply_pattern(%s,%s)\n",
              newname,
              p->pattern,
              filename);
              
   return newname;
}

/*
   get_pattern() - retrieve the pattern as input
   =============
*/
char *get_pattern(PCB *p)
{
   return p->pattern;
}

/*
   Private functions
   =================

   clear_error() - chear the error flag.
   =============
*/
void clear_error(PCB *p)
{
   if (p->error)
   {
      release_space(p->error);
      p->error = NULL;
   }
}

/*
   dump_pattern() - diagnostic output of a pattern
   ==============
*/
void dump_pattern(PCB *p)
{
   int         i;
   ELEMENT     *e;
   
   fprintf(stderr, " Pattern: %s\n", p->pattern);
   fprintf(stderr, "   Error: %s\n", ( p->error ? p->error : "none") );
   fprintf(stderr, "    Date: ");
   if (p->date)
      fprintf(stderr,
              "%04d-%02d-%02d %02d:%02d:%02d\n",
              p->date->tm_year + 1900,
              p->date->tm_mon + 1,
              p->date->tm_mday,
              p->date->tm_hour,
              p->date->tm_min,
              p->date->tm_sec);
   else
      fprintf(stderr, "not set\n");
              
   fprintf(stderr, "Elements: %d\n", p->elements);
   for (i = 0; i < p->elements; i++)
   {
      e = p->e[i];
      fprintf(stderr, "     %3d: %c ", i, e->code); 
      switch(e->code)
      {
         case CENTURY:
         case YEAR:
         case MONTH:
         case DAY:
         case HOUR:
         case MINUTE:
         case SECOND:
               fprintf(stderr, "\n");
               break;
               
         case PREFIX:
         case SUFFIX:
               fprintf(stderr, "length %d\n", e->length);
               break;
               
         case FILENAME:
               fprintf(stderr,
                       "offset %d, length %d\n",
                       e->offset,
                       e->length);
               break;
               
         case TEXT:
         case EXTENSION:
            if (e->value)
               fprintf(stderr, "value [%s]\n", e->value);
            else
               fprintf(stderr, "(value undefined)\n");
            break;
               
         default:
            fprintf(stderr, "unknown pattern CODE: 0x%02x", e->code);
      }
   }
}

/*
   set_error_code() - create an error string describing a problem with
   ================   an element code.
*/
void set_error_code(PCB *p, char *fmt, char c)
{
   clear_error(p);
   p->error = (char*)get_space(strlen(fmt) + 1);
   sprintf(p->error, fmt, c);
}

/*
   set_error_num() - build error string for a known element code and number
   ===============
*/
void set_error_num(PCB *p, char e, char *s, int n)
{
   clear_error(p);
   p->error = (char*)get_space(strlen(s) + 7);
   sprintf(p->error, "element %c - %s %d", e, s, n);
} 

/*
   set_error_text() - build error string for a known element code.
   ================
*/
void set_error_text(PCB *p, char e, char *s, char c)
{
   clear_error(p);
   p->error = (char*)get_space(strlen(s) + 6);
   sprintf(p->error, "%c %s %c", e, s, c);
} 

/*
   set_sys_error() - build an error string after a system error ocurrs
   ===============
*/
void set_sys_error(PCB *p, char *s, char *e)
{
   clear_error(p);
   p->error = (char*)get_space(strlen(s) + strlen(e) + 3);
   sprintf(p->error, "%s: %s", s, e);
} 
